/*
 * _cubrid - low level Python module for CUBRID
 *
 * copyright : (C) 2006 by Kang, Dong-Wan <dwkang@cubrid.com>
 * homepage  : http://pycubrid.kldp.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef _WINDOWS
#pragma warning(disable:4996) /* _CRT_SECURE_NO_DEPRECATE for VS.NET2005 */
#endif

#include "Python.h" 
#include "structmember.h"
#include "cas_cci.h"

#define MAX_STR_LEN	255

#define CUBRID_ER_INVALID_SQL_TYPE              -2002
#define CUBRID_ER_CANNOT_GET_COLUMN_INFO        -2003
#define CUBRID_ER_INIT_ARRAY_FAIL               -2004
#define CUBRID_ER_UNKNOWN_TYPE                  -2005
#define CUBRID_ER_INVALID_PARAM                 -2006
#define CUBRID_ER_INVALID_ARRAY_TYPE            -2007
#define CUBRID_ER_NOT_SUPPORTED_TYPE            -2008
#define CUBRID_ER_OPEN_FILE                     -2009
#define CUBRID_ER_CREATE_TEMP_FILE              -2010
#define CUBRID_ER_TRANSFER_FAIL                 -2011

char *cci_client_name = "CCI"; /* for CCI library */

static PyObject	*cubrid_error;
static PyObject	*cubrid_warning;
static PyObject	*cubrid_interface_error;
static PyObject	*cubrid_database_error;
static PyObject	*cubrid_internal_error;
static PyObject	*cubrid_operational_error;
static PyObject	*cubrid_programming_error;
static PyObject	*cubrid_integrity_error;
static PyObject	*cubrid_data_error;
static PyObject	*cubrid_not_supported_error;

extern PyTypeObject connection_object_type;
extern PyTypeObject cursor_object_type;
 
typedef struct _connection_object {
  PyObject_HEAD
  int                   handle;
  char                  host[MAX_STR_LEN];
  int                   port;
  char                  db[MAX_STR_LEN];
  char                  user[MAX_STR_LEN];
} Connection_Object;

typedef struct _cursor_object {
  PyObject_HEAD
  int                   connection;
  int                   handle;
  int                   affected_rows;
  int                   async_mode;
  int                   col_count;
  int                   rowCount;
  T_CCI_COL_INFO        *col_info;
  PyObject				*description;
  char                  *last_stmt;
} Cursor_Object;

static struct _error_message {
  int   err;
  char* msg;
} cubrid_err_msgs[] = {
{-1,      "CUBRID database error"},
{-2,      "Invalid connection handle"},
{-3,      "Memory allocation error"},
{-4,      "Communication error"},
{-5,      "No more data"},
{-6,      "Unknown transaction type"},
{-7,      "Invalid string parameter"},
{-8,      "Type conversion error"},
{-9,      "Parameter binding error"},
{-10,     "Invalid type"},
{-11,     "Parameter binding error"},
{-12,     "Invalid database parameter name"},
{-13,     "Invalid column index"},
{-14,     "Invalid schema type"},
{-15,     "File open error"},
{-16,     "Connection error"},
{-17,     "Connection handle creation error"},
{-18,     "Invalid request handle"},
{-19,     "Invalid cursor position"},
{-20,     "Object is not valid"},
{-21,     "CAS error"},
{-22,     "Unknown host name"},
{-99,     "Not implemented"},
{-1000,   "Database connection error"},
{-1002,   "Memory allocation error"},
{-1003,   "Communication error"},
{-1004,   "Invalid argument"},
{-1005,   "Unknown transaction type"},
{-1007,   "Parameter binding error"},
{-1008,   "Parameter binding error"},
{-1009,   "Cannot make DB_VALUE"},
{-1010,   "Type conversion error"},
{-1011,   "Invalid database parameter name"},
{-1012,   "No more data"},
{-1013,   "Object is not valid"},
{-1014,   "File open error"},
{-1015,   "Invalid schema type"},
{-1016,   "Version mismatch"},
{-1017,   "Cannot process the request. Try again later."},
{-1018,   "Authorization error"},
{-1020,   "The attribute domain must be the set type."},
{-1021,   "The domain of a set must be the same data type."},
{-2001,   "Memory allocation error"},
{-2002,   "Invalid API call"},
{-2003,   "Cannot get column info"},
{-2004,   "Array initializing error"},
{-2005,   "Unknown column type"},
{-2006,   "Invalid parameter"},
{-2007,   "Invalid array type"},
{-2008,   "Invalid type"},
{-2009,   "File open error"},
{-2010,   "Temporary file open error"},
{-2011,   "Glo transfering error"},
{0, ""}
};

int 
get_error_msg(int err_code, char **err_msg)
{
  int i;

  for(i = 0; ; i++) {
    if (!cubrid_err_msgs[i].err) break;
    if (cubrid_err_msgs[i].err == err_code) {
      *err_msg = cubrid_err_msgs[i].msg;
      return 0;
    }
  }
  return -1;
}

static PyObject *
handle_error(int e, T_CCI_ERROR *error)
{
  PyObject  *t;
  int       err_code;
  char      msg[1024];
  char      *err_msg = NULL, *facility_msg;

  if (e == CCI_ER_DBMS) {
    facility_msg = "DBMS";
    if (error) {
      err_code = error->err_code;
      err_msg = error->err_msg;
    } else {
      err_code = 0;
      err_msg = "Unknown DBMS Error";
    }
  } else {
    if (get_error_msg(e, &err_msg) < 0) {
      strcpy(err_msg, "Unknown Error");
    }
    err_code = e;

    if (e > -1000) {
      facility_msg = "CCI";
    } else if (e > -2000) {
      facility_msg = "CAS";
    } else if (e > -2000) {
      facility_msg = "CLIENT";
    } else {
      facility_msg = "UNKNOWN";
    }
  }

  sprintf(msg, "ERROR: %s, %d, %s", facility_msg, err_code, err_msg);

  if (!(t = PyTuple_New(2))) return NULL;
  PyTuple_SetItem(t, 0, PyInt_FromLong((long)e));
  PyTuple_SetItem(t, 1, PyString_FromString(msg));

  PyErr_SetObject(cubrid_error, t);
  Py_DECREF(t);

  return NULL;
}

static char cubrid_connect__doc__[] =
"Returns a CUBRID connection object.\n\n\
host\n\
  string, host to connect\n\n\
port\n\
  integer, TCP/IP port to connect to\n\n\
db\n\
  string, database to use\n\n\
user\n\
  string, user to connect as\n\n\
passwd\n\
  string, password to use\n\n";
 
static PyObject *
cubrid_connect(PyObject *self, PyObject *args, PyObject *kwargs)
{
  Connection_Object *c = NULL;
  static char *kwList[] = {"host", "port", "db", "user", "password", NULL};
  char *host	= "localhost";
  int  port	    = 30000;
  char *db	    = "demodb";
  char *user	= "public";
  char *passwd	= "";
  int  res, res2;
  char db_ver[16];
  T_CCI_ERROR error;

  if (!PyArg_ParseTupleAndKeywords(
        args, kwargs, "|sisss", kwList, &host, &port, &db, &user, &passwd)) {
    return NULL;
  }

  if (!(c = PyObject_NEW(Connection_Object, &connection_object_type))) {
    return PyErr_NoMemory();
  }

  c->handle = 0;
  c->host[0] = '\0';
  c->port = 0;
  c->db[0] = '\0';
  c->user[0] = '\0';

  res = cci_connect(host, port, db, user, passwd);

  if (res < 0) {
    Py_DECREF(c);
    return handle_error(res, NULL);
  }

  res2 = cci_get_db_version(res, db_ver, sizeof(db_ver));
  if (res2 < 0) {
    Py_DECREF(c);
	cci_disconnect(res, &error);
    return handle_error(res2, NULL);
  }

  res2 = cci_end_tran(res, CCI_TRAN_COMMIT, &error);
  if (res2 < 0) {
    Py_DECREF(c);
	cci_disconnect(res, &error);
    return handle_error(res2, &error);
  }

  c->handle = res;
  strcpy(c->host, host);
  c->port = port;
  strcpy(c->db, db);
  strcpy(c->user, user);

  return (PyObject *)c;
} 

static PyObject *
connection_object_close(Connection_Object *self, PyObject *args)
{
  if (self->handle) {
    T_CCI_ERROR error;
    cci_disconnect(self->handle, &error);
    self->handle = 0;
  }

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
connection_object_cursor(Connection_Object *self, PyObject *args)
{
  Cursor_Object *c;

  if (!self->handle) {
    handle_error(CCI_ER_CON_HANDLE, NULL);	
    return NULL;
  }

  if (!(c = PyObject_NEW(Cursor_Object, &cursor_object_type))) {
    return PyErr_NoMemory();
  }

  c->handle = 0;
  c->connection = self->handle;
  Py_INCREF(Py_None);
  c->description = Py_None;
  c->rowCount = -1;
  c->last_stmt = NULL;

  return (PyObject *)c;
}

static PyObject *
connection_object_end_tran(Connection_Object *self, int type)
{
  int res;
  T_CCI_ERROR error;

  res = cci_end_tran(self->handle, type, &error);
  if (res < 0){
    return handle_error(res, &error);
  }

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
connection_object_commit(Connection_Object *self, PyObject *args)
{
  return connection_object_end_tran(self, CCI_TRAN_COMMIT);
}

static PyObject *
connection_object_rollback(Connection_Object *self, PyObject *args)
{
  return connection_object_end_tran(self, CCI_TRAN_ROLLBACK);
}

static void
connection_object_dealloc(Connection_Object *self)
{
  PyObject *o;

  o = connection_object_close(self, NULL);
  Py_XDECREF(o);

  PyObject_DEL(self);
}

static PyObject *
connection_object_repr(Connection_Object *self)
{
  char buf[1024];
  if (self->handle) {
    sprintf(buf,
        "<open CUBRID connection at %lx, "
        "host: %s, "
        "port: %d, "
        "db: %s "
        "user: %s"
        ">",
        (long)self,
        self->host,
        self->port,
        self->db,
        self->user
        );
  } else {
    sprintf(buf, "<closed connection at %lx>", (long)self);
  }

  return PyString_FromString(buf);
}

static PyMethodDef connection_object_methods[] = {
  {
    "close",
    (PyCFunction) connection_object_close,
    METH_VARARGS,
    "close connection"
  },
  {
    "cursor",
    (PyCFunction) connection_object_cursor,
    METH_VARARGS,
    "get cursor"
  },
  {
    "commit",
    (PyCFunction) connection_object_commit,
    METH_VARARGS,
    "commit transaction"
  },
  {
    "rollback",
    (PyCFunction) connection_object_rollback,
    METH_VARARGS,
    "rollback transaction"
  },
  {NULL, NULL} /* sentinel */
};

static PyObject *
connection_object_getattr(Connection_Object *self, char *name)
{
  return Py_FindMethod(connection_object_methods, (PyObject *)self, name);
}

PyTypeObject connection_object_type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "_cubrid.connection", /* (char *)tp_name For printing */
    sizeof(Connection_Object),
    0,
    (destructor)connection_object_dealloc, /* tp_dealloc */
    0, /*tp_print*/
    (getattrfunc)connection_object_getattr, /* tp_getattr */
    0, /*(setattrfunc)connection_object_setattr,*/ /* tp_setattr */
    0, /*tp_compare*/
    (reprfunc)connection_object_repr, /* tp_repr */

    /* Method suites for standard classes */

    0, /* (PyNumberMethods *) tp_as_number */
    0, /* (PySequenceMethods *) tp_as_sequence */
    0, /* (PyMappingMethods *) tp_as_mapping */

    /* More standard operations (here for binary compatibility) */

    0, /* (hashfunc) tp_hash */
    0, /* (ternaryfunc) tp_call */
    0, /* (reprfunc) tp_str */
    0, /* (getattrofunc) tp_getattro */
    0, /* (setattrofunc) tp_setattro */

    /* Functions to access object as input/output buffer */
    0, /* (PyBufferProcs *) tp_as_buffer */

    /* Flags to define presence of optional/expanded features */
    Py_TPFLAGS_DEFAULT, /* | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,*/
    cubrid_connect__doc__, /* (char *) tp_doc Documentation string */

    /* Assigned meaning in release 2.0 */
    /* call function for all accessible objects */
    0, /*(traverseproc) connection_object_traverse,*/ /* tp_traverse */
    /* delete references to contained objects */
    0, /*(inquiry) connection_object_clear,*/ /* tp_clear */

    /* Assigned meaning in release 2.1 */
    /* rich comparisons */
    0, /* (richcmpfunc) tp_richcompare */
    /* weak reference enabler */
    0, /* (long) tp_weaklistoffset */

    /* Added in release 2.2 */
    /* Iterators */
    0, /* (getiterfunc) tp_iter */
    0, /* (iternextfunc) tp_iternext */

    /* Attribute descriptor and subclassing stuff */
    (struct PyMethodDef *)connection_object_methods, /* tp_methods */
    0, /*(MyMemberlist(*))connection_object_memberlist,*/ /* tp_members */
    0, /* (struct getsetlist *) tp_getset; */
    0, /* (struct _typeobject *) tp_base; */
    0, /* (PyObject *) tp_dict */
    0, /* (descrgetfunc) tp_descr_get */
    0, /* (descrsetfunc) tp_descr_set */
    0, /* (long) tp_dictoffset */
    0, /*(initproc)connection_object_Initialize,*/ /* tp_init */
    NULL, /* tp_alloc */
    NULL, /* tp_new */
    NULL, /* tp_free Low-level free-memory routine */ 
    0, /* (PyObject *) tp_bases */
    0, /* (PyObject *) tp_mro method resolution order */
    0, /* (PyObject *) tp_defined */
};

static void
cursor_object_reset(Cursor_Object *self)
{
  int res;

  if (self->handle) {
    res = cci_close_req_handle(self->handle);
    self->handle = 0;

	if (self->description) {
	  Py_DECREF(self->description);
	  self->description = NULL;
	}

    if (self->last_stmt) {
      PyMem_Free(self->last_stmt);
	  self->last_stmt = NULL;
    }
  }
}

static PyObject *
cursor_object_close(Cursor_Object *self, PyObject *args)
{
  cursor_object_reset(self);
  Py_INCREF(Py_None);
  return Py_None;
}

static int
cursor_object_set_description(Cursor_Object *self)
{
  PyObject	*desc;
  int i;
  char colName[128];
  int datatype, precision, scale, nullable;

  if (self->col_count == 0) {
    Py_XDECREF(self->description);
    self->description = PyTuple_New(0);
    return 1;
  }

  desc = (PyObject *) PyTuple_New(self->col_count);

  for ( i = 0 ; i < self->col_count ; i++ ) {
    PyObject *item;

    item = PyTuple_New(7);

    strcpy(colName, CCI_GET_RESULT_INFO_NAME(self->col_info, i+1));
    precision = CCI_GET_RESULT_INFO_PRECISION(self->col_info, i+1);
    scale     = CCI_GET_RESULT_INFO_SCALE(self->col_info, i+1);
    nullable  = CCI_GET_RESULT_INFO_IS_NON_NULL(self->col_info, i+1);

    datatype  = CCI_GET_RESULT_INFO_TYPE(self->col_info, i+1);
    /*
     * char 			1
     * string,varchar 	2
     * nchar 			3
     * varnchar			4
     * bit	 			5
     * varbit			6
     * numeric	 		7
     * int	 			8
     * short 			9
     * monetary 		10
     * float 			11
     * double 			12
     * date 			13
     * time 			14
     * timestamp		15
     * object 			19
     * set				32
     * multiset			64
     * sequence			96
     */
     
    PyTuple_SetItem(item, 0, PyString_FromString(colName));
    PyTuple_SetItem(item, 1, PyInt_FromLong(datatype));
    PyTuple_SetItem(item, 2, PyInt_FromLong(0));
    PyTuple_SetItem(item, 3, PyInt_FromLong(0));
    PyTuple_SetItem(item, 4, PyInt_FromLong(precision));
    PyTuple_SetItem(item, 5, PyInt_FromLong(scale));
    PyTuple_SetItem(item, 6, PyInt_FromLong(nullable));

    PyTuple_SetItem(desc, i, item);
  }

  Py_XDECREF(self->description);
  self->description = desc;

  return 1;
}

static int
cursor_object_bind_param(Cursor_Object *self, PyObject *params, T_CCI_PARAM_INFO *param_info)
{
  int i, res;
  PyObject	*paramVal;
  int param_cnt;
  char *str_val;

  param_cnt = PyTuple_Size(params);

  for(i = 0; i < param_cnt; i++) {
    paramVal = PyTuple_GetItem(params, i);
    if (PyString_Check(paramVal)) {
      str_val = PyString_AsString(paramVal);
      res = cci_bind_param(self->handle, i + 1, CCI_A_TYPE_STR, str_val, CCI_U_TYPE_STRING, 0);
    }
    else {
      res = CUBRID_ER_UNKNOWN_TYPE;
    }

    if (res < 0) {
      return res;
    }
  }

  return 0;
}

static PyObject *
cursor_object_execute(Cursor_Object *self, PyObject *args)
{
  T_CCI_ERROR error;
  T_CCI_COL_INFO *res_col_info;
  T_CCI_SQLX_CMD res_sql_type;
  T_CCI_PARAM_INFO *res_param_info = NULL;
  char 		*stmt = "";
  PyObject	*params = NULL;
  int 		res, res_col_count;

  if (!PyArg_ParseTuple(args, "s|O", &stmt, &params)) {
    return NULL;
  }

  if (!self->last_stmt || strcmp(self->last_stmt, stmt) != 0) {
    cursor_object_reset(self);

    res = cci_prepare(self->connection, stmt, 0 /*CCI_PREPARE_INCLUDE_OID*/, &error);
    if (res < 0) {
      return handle_error(res, &error);
    }
    self->handle = res;
    if (self->last_stmt) {
      PyMem_Free(self->last_stmt);
	  self->last_stmt = NULL;
    }
    self->last_stmt = PyMem_Malloc(strlen(stmt) + 1);
    strcpy(self->last_stmt, stmt);
  }

  if (params) {
    res = cursor_object_bind_param(self, params, res_param_info);
    if (res < 0) {
      return handle_error(res, NULL);
    } 
  }

  res = cci_execute(self->handle, 0/*CCI_EXEC_ASYNC*/, 0, &error);
  if (res < 0) {
    return handle_error(res, &error);
  }

  res_col_info = cci_get_result_info(self->handle, &res_sql_type, &res_col_count);
  if (!res_col_info) {
    return handle_error(CUBRID_ER_CANNOT_GET_COLUMN_INFO, NULL);
  }

  self->col_info = res_col_info;
  self->col_count = res_col_count;
  self->async_mode = 2;

  switch(res_sql_type) {
    case SQLX_CMD_SELECT:
      self->rowCount = res;
      break;
    case SQLX_CMD_INSERT:
    case SQLX_CMD_UPDATE:
    case SQLX_CMD_DELETE:
      self->affected_rows = res;
      break;
    case SQLX_CMD_CALL:
      self->rowCount = res;

    default:
      break;
  }

  if (res_sql_type == SQLX_CMD_SELECT) {
    cursor_object_set_description(self);

    res = cci_cursor(self->handle, 1, CCI_CURSOR_CURRENT, &error);
    if (res < 0 && res != CCI_ER_NO_MORE_DATA) {
      return handle_error(res, &error);
    }
  }

  return Py_BuildValue("i", self->rowCount);
}

/* DB type to Python type mapping
 * 
 * int, short 				-> Integer
 * float, double, numeric 	-> Float
 * another type				-> String
 */
static PyObject *
cursor_object_dbval_to_pyvalue(Cursor_Object *self, int type, int index)
{
  int res, ind;
  PyObject *val, *tmpval;
  char *res_buf;
  int int_val;

  switch (type) {
    case CCI_U_TYPE_INT:
    case CCI_U_TYPE_SHORT:
      res = cci_get_data(self->handle, index, CCI_A_TYPE_INT, &int_val, &ind);
      if (res < 0) {
        return handle_error(res, NULL);
      }
	  if (ind < 0) {
		Py_INCREF(Py_None);
		val = Py_None;
	  } else {
        val = PyInt_FromLong(int_val);
      }
      break;

    case CCI_U_TYPE_FLOAT:
    case CCI_U_TYPE_DOUBLE:
    case CCI_U_TYPE_NUMERIC:
      res = cci_get_data(self->handle, index, CCI_A_TYPE_STR, &res_buf, &ind);
      if (res < 0) {
        return handle_error(res, NULL);
      }
	  if (ind < 0) {
		Py_INCREF(Py_None);
		val = Py_None;
	  } else {
        tmpval = PyString_FromString(res_buf);
        val = PyFloat_FromString(tmpval, NULL);
        Py_DECREF(tmpval);
      }
      break;

    default:
	  res = cci_get_data(self->handle, index, CCI_A_TYPE_STR, &res_buf, &ind);
	  if (res < 0) {
		return handle_error(res, NULL);
	  }
	  if (ind < 0) {
		Py_INCREF(Py_None);
		val = Py_None;
	  } else {
		val = PyString_FromString(res_buf);
	  }
      break;
  }

  return val;
}

/* Collection(set, multiset, sequence) 	-> List, 
 * Collection' item 					-> String 
 */
static PyObject *
cursor_object_dbset_to_pyvalue(Cursor_Object *self, int index)
{
  int i, res, ind;
  PyObject *val;
  T_CCI_SET set;
  int set_size;
  PyObject *e;
  char *buffer;

  res = cci_get_data(self->handle, index, CCI_A_TYPE_SET, &set, &ind);
  if (res < 0) {
    return handle_error(res, NULL);
  }

  if (set == NULL) {
    Py_INCREF(Py_None);
    return Py_None;
  }
  
  set_size = cci_set_size(set);
  val = PyList_New(set_size);

  for (i = 0; i < set_size; i++) {
    res = cci_set_get(set, i + 1, CCI_A_TYPE_STR, &buffer, &ind);
    if (res < 0) {
      return handle_error(res, NULL);
    }
    e =  PyString_FromString(buffer);
    PyList_SetItem(val, i, e);
  } 
  cci_set_free(set); 

  return val;
}

static PyObject *
cursor_object_fetch_a_row(Cursor_Object *self)
{
  int i, type;
  PyObject *row; 
  PyObject *val;

  row = PyList_New(self->col_count);

  for (i = 0; i < self->col_count; i++) {
    type = CCI_GET_RESULT_INFO_TYPE(self->col_info, i + 1);

    if (CCI_IS_COLLECTION_TYPE(type)) {
      val = cursor_object_dbset_to_pyvalue(self, i + 1);
    } else {
      val = cursor_object_dbval_to_pyvalue(self, type, i + 1);
    }
    PyList_SetItem(row, i, val);
  }

  return row;
}

static PyObject *
cursor_object_fetch(Cursor_Object *self, PyObject *args)
{
  int res;
  T_CCI_ERROR error;
  PyObject *row;

  res = cci_cursor(self->handle, 0, CCI_CURSOR_CURRENT, &error);
  if (res == CCI_ER_NO_MORE_DATA) {
    Py_INCREF(Py_None);
    return Py_None;
  } else if (res < 0) {
    return handle_error(res, &error);
  }

  res = cci_fetch(self->handle, &error);
  if (res < 0) {
    return handle_error(res, &error);
  }

  row = cursor_object_fetch_a_row(self);

  res = cci_cursor(self->handle, 1, CCI_CURSOR_CURRENT, &error);
  if (res < 0 && res != CCI_ER_NO_MORE_DATA) {
    return handle_error(res, &error);
  }

  return row;
}

static struct PyMemberDef cursor_object_members[] = {
  {
    "description", 
    T_OBJECT, 
    offsetof(Cursor_Object, description), 
    0, 
    "connection"
  },
  { 
    "rowcount", 
    T_INT, 
    offsetof(Cursor_Object, rowCount), 
    0, 
    "row count" 
  },
  { NULL }
};

static PyMethodDef cursor_object_methods[] = {
  {
    "close",
    (PyCFunction) cursor_object_close,
    METH_VARARGS, 
    "close cursor"
  },
  {
    "execute",
    (PyCFunction) cursor_object_execute,
    METH_VARARGS,
    "execute query"
  },
  {
    "fetch",
    (PyCFunction) cursor_object_fetch,
    METH_VARARGS,
    "fetch result"
  },
  {NULL, NULL} /* sentinel */
};

static void
cursor_object_dealloc(Cursor_Object *self)
{
  cursor_object_reset(self);
  PyObject_DEL(self);
}

static PyObject *
cursor_object_getattr(Cursor_Object *self, char *name)
{
  PyObject *res;
  struct PyMemberDef *l;

  res = Py_FindMethod(cursor_object_methods, (PyObject *)self, name);
  if (res != NULL)
    return res;
  PyErr_Clear();

  for (l = cursor_object_members; l->name != NULL; l++) {
    if (strcmp(l->name, name) == 0)
      return PyMember_GetOne((char *)self, l);
  }
  PyErr_SetString(PyExc_AttributeError, name);
  return NULL;
}

static char cursor_object__doc__[] = "";

PyTypeObject cursor_object_type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "_cubrid.cursor",
    sizeof(Cursor_Object),
    0,
    (destructor)cursor_object_dealloc, /* tp_dealloc */
    0, /*tp_print*/
    (getattrfunc)cursor_object_getattr, /* tp_getattr */
    0, /*(setattrfunc)cursor_object_setattr,*/ /* tp_setattr */
    0, /*tp_compare*/
    0, /*(reprfunc)cursor_object_repr,*/ /* tp_repr */

    /* Method suites for standard classes */

    0, /* (PyNumberMethods *) tp_as_number */
    0, /* (PySequenceMethods *) tp_as_sequence */
    0, /* (PyMappingMethods *) tp_as_mapping */

    /* More standard operations (here for binary compatibility) */

    0, /* (hashfunc) tp_hash */
    0, /* (ternaryfunc) tp_call */
    0, /* (reprfunc) tp_str */
    0, /* (getattrofunc) tp_getattro */
    0, /* (setattrofunc) tp_setattro */

    /* Functions to access object as input/output buffer */
    0, /* (PyBufferProcs *) tp_as_buffer */

    /* Flags to define presence of optional/expanded features */
    Py_TPFLAGS_DEFAULT, /* | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,*/

    cursor_object__doc__, /* (char *) tp_doc Documentation string */

    /* Assigned meaning in release 2.0 */
    /* call function for all accessible objects */
    0, /*(traverseproc) cursor_object_traverse,*/ /* tp_traverse */
    /* delete references to contained objects */
    0, /*(inquiry) cursor_object_clear,*/ /* tp_clear */

    /* Assigned meaning in release 2.1 */
    /* rich comparisons */
    0, /* (richcmpfunc) tp_richcompare */

    /* weak reference enabler */
    0, /* (long) tp_weaklistoffset */

    /* Added in release 2.2 */
    /* Iterators */
    0, /* (getiterfunc) tp_iter */
    0, /* (iternextfunc) tp_iternext */

    /* Attribute descriptor and subclassing stuff */
    (struct PyMethodDef *) cursor_object_methods, /* tp_methods */
    (struct PyMemberDef *) cursor_object_members, /*tp_members */
    0, /* (struct getsetlist *) tp_getset; */
    0, /* (struct _typeobject *) tp_base; */
    0, /* (PyObject *) tp_dict */
    0, /* (descrgetfunc) tp_descr_get */
    0, /* (descrsetfunc) tp_descr_set */
    0, /* (long) tp_dictoffset */
    0, /*(initproc)cursor_object_Initialize,*/ /* tp_init */
    NULL, /* tp_alloc */
    NULL, /* tp_new */
    NULL, /* tp_free Low-level free-memory routine */
    0, /* (PyObject *) tp_bases */
    0, /* (PyObject *) tp_mro method resolution order */
    0, /* (PyObject *) tp_defined */
};
 
static struct PyMethodDef _cubrid_methods[] = { 
  {
    "connect", 
    (PyCFunction) cubrid_connect, 
    METH_VARARGS | METH_KEYWORDS, 
    "connect to CUBRID server"
  }, 
  {NULL, NULL} 
};             

void 
init_exceptions(PyObject *dict)
{
  cubrid_error = PyErr_NewException("_cubrid.Error", PyExc_StandardError, NULL);
  PyDict_SetItemString(dict, "Error", cubrid_error);

  cubrid_warning = PyErr_NewException("_cubrid.Warning", PyExc_StandardError, NULL);
  PyDict_SetItemString(dict, "Warning", cubrid_warning);

  cubrid_interface_error = PyErr_NewException("_cubrid.InterfaceError", cubrid_error, NULL);
  PyDict_SetItemString(dict, "InterfaceError",cubrid_interface_error);

  cubrid_database_error = PyErr_NewException("_cubrid.DatabaseError", cubrid_error, NULL);
  PyDict_SetItemString(dict, "DatabaseError", cubrid_database_error);

  cubrid_internal_error = PyErr_NewException("_cubrid.InternalError", cubrid_database_error, NULL);
  PyDict_SetItemString(dict, "InternalError", cubrid_internal_error);

  cubrid_operational_error = PyErr_NewException("_cubrid.OperationalError", cubrid_database_error, NULL);
  PyDict_SetItemString(dict, "OperationalError", cubrid_operational_error);

  cubrid_programming_error = PyErr_NewException("_cubrid.ProgrammingError", cubrid_database_error, NULL);
  PyDict_SetItemString(dict, "ProgrammingError", cubrid_programming_error);

  cubrid_integrity_error = PyErr_NewException("_cubrid.IntegrityError", cubrid_database_error, NULL);
  PyDict_SetItemString(dict, "IntegrityError", cubrid_integrity_error);

  cubrid_data_error = PyErr_NewException("_cubrid.DataError", cubrid_database_error, NULL);
  PyDict_SetItemString(dict, "DataError", cubrid_data_error);

  cubrid_not_supported_error = PyErr_NewException("_cubrid.NotSupportedError", cubrid_database_error, NULL);
  PyDict_SetItemString(dict, "NotSupportedError", cubrid_not_supported_error);
}

#define _CUBRID_VERSION_	"0.5.0"
static char _cubrid_doc_[] = "CUBRID API for Python";
 
void 
init_cubrid() 
{ 
  PyObject *dict, *module;

  module = Py_InitModule4("_cubrid", _cubrid_methods, _cubrid_doc_, (PyObject *)NULL, PYTHON_API_VERSION);

  if (!(dict = PyModule_GetDict(module))) {
    goto Error;
  }

  init_exceptions(dict);
  PyDict_SetItemString(dict, "__version__", PyString_FromString(_CUBRID_VERSION_));

Error:
  if (PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "_cubrid: init failure");
  }

  return;
} 

